from collections import namedtuple
from .pyte.graphics import FG_ANSI, BG_ANSI

fg2code = {name: str(key).encode("latin-1") for key, name in FG_ANSI.items()}
bg2code = {name: str(key).encode("latin-1") for key, name in BG_ANSI.items()}

RESET = b"\x1b[m"

ColorState = namedtuple("ColorState", ["bold", "blink", "fg", "bg"])

DEFAULT_COLOR = ColorState(False, False, "white", "black")

def char_to_color(char):
    fg = char.fg if char.fg != "default" else "white"
    bg = char.bg if char.bg != "default" else "black"
    if char.reverse:
        fg, bg = bg, fg
    return ColorState(char.bold, char.blink, fg, bg)        

def code_to_ansi(codes):
    return b"\x1b[" + b";".join(codes) + b"m"
    
def diff(color, other):
    """Rerturn an escape sequence that would transform color state from
    ``other`` to ``color``.
    """
    if color == DEFAULT_COLOR:
        return RESET # reset
    codes = []
    if (color.bold != other.bold and not color.bold or
            color.blink != other.blink and not color.blink):
        # reset
        codes.append(b"0")
        other = DEFAULT_COLOR
        
    if color.bold and color.bold != other.bold:
        codes.append(b"1")
    if color.blink and color.blink != other.blink:
        codes.append(b"5")
    if color.fg != other.fg:
        codes.append(fg2code[color.fg])
    if color.bg != other.bg:
        codes.append(bg2code[color.bg])
    return code_to_ansi(codes)
        
def colored_sequence(chars):
    color = DEFAULT_COLOR
    for c in chars:
        if c is None:
            yield b" "
            continue
        next_color = char_to_color(c)
        if color != next_color:
            yield diff(next_color, color)
            color = next_color
        yield c.data.encode("latin-1")
    if color != DEFAULT_COLOR:
        yield RESET
        
def chars_to_bytes(chars):
    """Convert a list of :class:`pyte.screens.Char` into ansi escape sequence.
    """
    return b"".join(colored_sequence(chars))



"""
A pyte screen that is allowed to draw unprintable characters.
"""

from wcwidth import wcwidth
from .pyte.screens import Screen, mo, unicodedata

class ByteScreen(Screen):
    def draw(self, data):
        data = data.translate(
            self.g1_charset if self.charset else self.g0_charset)

        for char in data:
            char_width = wcwidth(char)

            # If this was the last column in a line and auto wrap mode is
            # enabled, move the cursor to the beginning of the next line,
            # otherwise replace characters already displayed with newly
            # entered.
            if self.cursor.x == self.columns:
                if mo.DECAWM in self.mode:
                    self.dirty.add(self.cursor.y)
                    self.carriage_return()
                    self.linefeed()
                elif char_width > 0:
                    self.cursor.x -= char_width

            # If Insert mode is set, new characters move old characters to
            # the right, otherwise terminal is in Replace mode and new
            # characters replace old characters at cursor position.
            if mo.IRM in self.mode and char_width > 0:
                self.insert_characters(char_width)

            line = self.buffer[self.cursor.y]
            if char_width == 1 or char_width < 0:
                char_width = 1
                line[self.cursor.x] = self.cursor.attrs._replace(data=char)
            elif char_width == 2:
                # A two-cell character has a stub slot after it.
                line[self.cursor.x] = self.cursor.attrs._replace(data=char)
                if self.cursor.x + 1 < self.columns:
                    line[self.cursor.x + 1] = self.cursor.attrs \
                        ._replace(data="")
            elif char_width == 0 and unicodedata.combining(char):
                # A zero-cell character is combined with the previous
                # character either on this or preceeding line.
                if self.cursor.x:
                    last = line[self.cursor.x - 1]
                    normalized = unicodedata.normalize("NFC", last.data + char)
                    line[self.cursor.x - 1] = last._replace(data=normalized)
                elif self.cursor.y:
                    last = self.buffer[self.cursor.y - 1][self.columns - 1]
                    normalized = unicodedata.normalize("NFC", last.data + char)
                    self.buffer[self.cursor.y - 1][self.columns - 1] = \
                        last._replace(data=normalized)
            else:
                break  # Unprintable character or doesn't advance the cursor.

            # .. note:: We can't use :meth:`cursor_forward()`, because that
            #           way, we'll never know when to linefeed.
            if char_width > 0:
                self.cursor.x = min(self.cursor.x + char_width, self.columns)

        self.dirty.add(self.cursor.y)
        


import logging
import math
import re
from contextlib import contextmanager
from getpass import getpass

import uao
from paramiko.client import SSHClient, AutoAddPolicy

from .byte_screen import ByteScreen
from .byte_stream import ByteStream
from .article import match_foot, Article

uao.register_uao()

log = logging.getLogger(__name__)

RX_LAST_PAGE = re.compile(r"瀏覽.+?\(100%\)".encode("big5-uao"))

def is_no(text):
    return bool(re.match(r"\s*(n|no)\s*", text, re.I))
    
def parse_board_item(line):
    no, date, sender, title = (
        t.decode("big5-uao").strip()
        for t in (line[0:6], line[9:14], line[15:30], line[30:])
    )
    no = int(no.replace("●", "").strip())
    if title.startswith("轉"):
        title = "Fw:" + title[1:]
    elif title.startswith("◇"):
        title = title[2:]
    elif title.startswith("R:"):
        title = "Re:" + title[2:]
    return no, date, sender, title
    
@contextmanager
def ptt_login(user=None, password=None):
    with SSHClient() as client:
        client.set_missing_host_key_policy(AutoAddPolicy)
        client.connect("ptt.cc", username="bbs", password="")
        with client.invoke_shell() as channel:
            channel.settimeout(10)
            bot = PTTBot(channel)
            try:
                bot.login(user, password)
                yield bot
            except:
                log.info(
                    "uncaught error, here is the last screen:\n%s",
                    "\n".join(line.decode("big5-uao").rstrip() for line in bot.lines())
                )
                raise
    
class PTTBot:
    def __init__(self, channel):
        self.channel = channel
        self.screen = ByteScreen(80, 24)
        self.stream = ByteStream(self.screen)
        self.article_configured = False
        self.user = None
        
    def login(self, user=None, password=None):
        if not user:
            user = input("User: ")
        if not password:
            password = getpass()
        self.user = user
        self.unt("請輸入代號")
        
        log.info("start login")
        
        self.send(user + "\r" + password + "\r")
        def handle_login(data):
            if "刪除其他重複登入".encode("big5-uao") in data:
                self.send("n\r")
                
            if "密碼不對喔！".encode("big5-uao") in data:
                raise Exception("failed to login. Wrong password.")
        self.unt("按任意鍵繼續", on_data=handle_login)
        
        log.info("%s login success", user)
        
        self.send(" ")
        def handle_after_login(data):
            if "編輯器自動復原".encode("big5-uao") in data:
                raise Exception("failed to login. Unsaved article detected.")
                
            if "您要刪除以上錯誤嘗試的記錄嗎?".encode("big5-uao") in data:
                self.send("n\r")
                
            if re.search(r"您保存信件數目 \d+ 超出上限 \d+".encode("big5-uao"), data):
                self.send("qq")
                
            if "新看板，確定要加入我的最愛嗎".encode("big5-uao") in data:
                self.send("y\r")
        self.unt(self.detect("主功能表", 0), on_data=handle_after_login)
        log.info("enter main menu")
        
    def detect(self, needle, line_no):
        if needle.startswith("!"):
            reverse = True
            needle = needle[1:].encode("big5-uao")
        else:
            reverse = False
            needle = needle.encode("big5-uao")
        def callback(_data):
            if reverse:
                return needle not in self.get_line(line_no)
            return needle in self.get_line(line_no)
        return callback
        
    def unt(self, needle, on_data=None):
        if callable(needle):
            should_stop = needle
        else:
            def should_stop(data, test=needle.encode("big5-uao")):
                return test in data
                
        while True:
            data = self.channel.recv(math.inf)
            if on_data:
                on_data(data)
            self.stream.feed(data)
            if should_stop(data):
                return
        
    def send(self, data):
        pos = 0
        while True:
            sent = self.channel.send(data[pos:])
            pos += sent
            if pos >= len(data):
                return
        
    @contextmanager
    def enter_mail(self):
        self.send(b"\x1am")
        self.unt("郵件選單")
        log.info("get in the mail box")
        yield
        self.send("q")
        self.unt("主功能表")
        log.info("get out the mail box")
        
    def get_last_index(self):
        log.info("get last index")
        # set_trace()
        self.send("$h")
        last_index = None
        def on_data(data):
            nonlocal last_index
            if "呼叫小天使".encode("big5-uao") in data:
                no, *_args = parse_board_item(self.get_line(self.screen.cursor.y))
                last_index = no
        self.unt("呼叫小天使", on_data=on_data)
        self.send("q")
        log.info("get last index success: %s", last_index)
        return last_index
        
    def lines(self, raw=False):
        for i in range(self.screen.lines):
            if raw:
                yield self.get_raw_line(i)
            else:
                yield self.get_line(i)
                
    def get_raw_line(self, line_no):
        if line_no < 0:
            line_no += self.screen.lines
        return [self.screen.buffer[line_no][i] for i in range(self.screen.columns)]

    def get_line(self, line_no):
        chars = self.get_raw_line(line_no)
        return "".join(c.data for c in chars).encode("latin-1")
        
    def update_article_config(self):
        """Update article config. It is hard to work with articles containing
        long lines (column > 80).
        """
        log.info("update pmore config")
        self.article_configured = True
        self.send("o")
        self.unt(self.on_pmore_conf)
        self.send("wmlq")
        self.unt(self.on_col(0))
        
    def on_pmore_conf(self, _data):
        return "piaip's more: pmore 2007+ 設定選項".encode("big5-uao") in self.get_line(-9)
        
    def on_col(self, col_no):
        def callback(_data):
            foot = match_foot(self.get_line(-1))
            return foot and foot.col_start == col_no
        return callback
        
    def get_article(self, index):
        log.info("get %sth article", index)
        self.send(str(index))
        self.unt("跳至第幾項")
        self.send("\r")
        self.unt(self.detect("!跳至第幾項", -1))
        curr_line = self.get_line(self.screen.cursor.y)
        _no, date, sender, title = parse_board_item(curr_line)
        
        log.info("title: %s", title)
        
        self.send("x" + self.user + "\r")
        self.unt(self.detect("標  題:", 2))
        title = self.get_line(2)[8:].strip()[:-5].strip().decode("big5-uao")
        self.send("n\r\x18a\r\r")
        self.unt(self.detect("郵件選單", 0))
        
        article = Article(date, sender, title)
        
        self.send("\r")
        
        is_animated = False
        def handle_animated(data):
            nonlocal is_animated
            if "這份文件是可播放的文字動畫".encode("big5-uao") in data:
                log.info("skip animation")
                self.send("n")
                is_animated = True
        self.unt(self.on_col(0), on_data=handle_animated)
        log.info("enter the article. is_animated=%s", is_animated)
        
        if is_animated:
            self.send("hq")
            self.unt(self.on_col(0))
            log.info("refresh animation page")
            
        if not self.article_configured:
            self.update_article_config()
            
        log.info("start collecting body")
        while True:
            screen = article.add_screen(self.lines(raw=True))
            log.info("add screen %s~%s", screen.line_start, screen.line_end)
            
            indent = 0
            while any(line.right_truncated for line in screen.lines):
                truncated_lines = set(line.line_no for line in screen.lines if line.right_truncated)
            
                log.info("has truncated right")
                indent += 1
                self.send(">")
                if screen.col_start == 0:
                    # the first indent is shorter
                    next_col = 7
                else:
                    next_col = screen.col_start + 8
                self.unt(self.on_col(next_col))
                screen = article.add_screen(
                    self.lines(raw=True),
                    skip_line=lambda line: line.line_no not in truncated_lines
                )
                log.info("move right to col %s", screen.col_start)
                # if screen.col_start == 136:
                    # set_trace()
                
            log.info("max indent %s", indent)
            if indent:
                self.send("<" * indent)
                self.unt(self.on_col(0))
                log.info("back to first col")
            
            if self.on_last_page():
                break
                
            self.send(" ")
            self.unt(lambda _data: (
                self.on_last_page() or self.on_line(screen.line_start + self.screen.lines - 2)
            ))
        self.send("q")
        
        log.info("get article success")
        return article
    
    def on_last_page(self):
        return RX_LAST_PAGE.search(self.get_line(-1))
        
    def on_line(self, line_no):
        foot = match_foot(self.get_line(-1))
        return foot and foot.line_start == line_no



import re
import collections
import uao

from .ansi import chars_to_bytes

uao.register_uao()

RX_FOOT = re.compile(r"(?:(\d+)~(\d+)\s*欄.+?)?(\d+)~(\d+)\s*行".encode("big5-uao"))

def is_default(char):
    return not char.bold and char.fg == "default" and char.bg == "default"
    
def is_truncated(char):
    return char.bold and (char.data == "<" or char.data == ">")
    
class ArticleScreenLine:
    def __init__(self, line, line_no, col_start):
        skip_start = 0
        if is_truncated(line[0]):
            if line[1].data == "?":
                skip_start = 2
            else:
                skip_start = 1
                
        skip_end = 0
        if is_truncated(line[78]) and is_default(line[79]):
            if line[77].data == "?":
                skip_end = 3
            else:
                skip_end = 2
                
        self.line_no = line_no
        self.col_no = col_start + skip_start
        self.chars = line[skip_start:len(line) - skip_end]
        self.left_truncated = bool(skip_start)
        self.right_truncated = bool(skip_end)
        
Foot = collections.namedtuple("Foot", ["col_start", "col_end", "line_start", "line_end"])
        
def match_foot(s):
    match = RX_FOOT.search(s)
    if not match:
        return None
    col_start, col_end, line_start, line_end = match.groups()
    
    col_start = int(col_start) - 2 if col_start is not None else 0
    col_end = int(col_end) if col_end is not None else 78
    line_start = int(line_start) - 1
    line_end = int(line_end)
    return Foot(col_start, col_end, line_start, line_end)
        
class ArticleScreen:
    def __init__(self, lines):
        lines = list(lines)
        foot = match_foot("".join(c.data for c in lines[-1]).encode("latin-1"))
        
        self.line_start = foot.line_start
        self.line_end = foot.line_end
        self.col_start = foot.col_start
        self.col_end = foot.col_end
        
        self.lines = [
            ArticleScreenLine(line, line_no, self.col_start)
            for line_no, line in enumerate(lines[:-1], self.line_start)
        ]
        
class Article:
    """Article composer. Compose multiple screens into an article."""
    def __init__(self, date, sender, title):
        self.date = date
        self.sender = sender
        self.title = title
        self.lines = []
        
    def draw_char(self, line_no, col_no, char):
        if col_no >= len(self.lines[line_no]):
            self.lines[line_no].extend([None] * (col_no - len(self.lines[line_no]) + 1))
                
        if self.lines[line_no][col_no] is None:
            self.lines[line_no][col_no] = char
        
    def draw_line(self, line):
        if line.line_no > len(self.lines):
            for _i in range(line.line_no - len(self.lines)):
                self.lines.append([])
                
        if line.line_no == len(self.lines) and line.col_no == 0:
            # no need to draw char
            self.lines.append(line.chars)
            return
            
        for col_no, char in enumerate(line.chars, line.col_no):
            self.draw_char(line.line_no, col_no, char)
        
    def add_screen(self, lines, skip_line=None):
        screen = ArticleScreen(lines)
        assert screen.line_start <= len(self.lines)
        for line in screen.lines:
            if skip_line and skip_line(line):
                continue
            self.draw_line(line)
        return screen
        
    def to_bytes(self):
        return b"\r\n".join(chars_to_bytes(l).rstrip() for l in self.lines)



# -*- coding: utf-8 -*-
"""
    pyte.compat
    ~~~~~~~~~~~

    Python version specific compatibility fixes.

    :copyright: (c) 2015-2017 by pyte authors and contributors,
                see AUTHORS for details.
    :license: LGPL, see LICENSE for more details.
"""

import sys

if sys.version_info[0] == 2:
    from future_builtins import map

    range = xrange
    str = unicode
    chr = unichr

    def pass_through_str(data):
        """Decode :func:`bytes` to :func:`str` using pass-through encoding."""
        return "".join(chr(ord(ch)) for ch in data)
else:
    from builtins import map, range, str, chr

    def pass_through_str(data):
        """Decode :func:`bytes` to :func:`str` using pass-through encoding."""
        return "".join(map(chr, data))
